#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `epages_provisioning` package."""

import os
import unittest
from datetime import datetime

from zeep.exceptions import ValidationError
from epages_provisioning import provisioning

# import logging
if os.environ.get('EP_TRACE', False):
    import logging
    import http.client as http_client
    http_client.HTTPConnection.debuglevel = 1
    logging.basicConfig()
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True


# check for our environment variables
envcheck = unittest.skipUnless(
    os.environ.get('EP_SERVER', False) and
    os.environ.get('EP_PROVIDER', False) and
    os.environ.get('EP_PASSWORD', False) and
    os.environ.get('EP_USERNAME', False), 'Environment variables not set'
)


@envcheck
class TestShopConfiguration(unittest.TestCase):
    """ Tests for `epages_provisioning` package.
    All tests require that the environment variables:

        EP_SERVER
        EP_PROVIDER
        EP_USERNAME
        EP_PASSWORD

    are set

    For example:

        export EP_SERVER=http://example.com/
        export EP_PROVIDER=Distributor
        export EP_USERNAME=admin
        export EP_PASSWORD=admin
        make test

    These tests also assume that the ePages service is a default installation
    with default shoptypes etc.

    Also if you want to see a full message trace say:

        export EP_TRACE=1

    and then run your tests

    Warning: these will take a while to run.
    """
    @classmethod
    def setUpClass(cls):
        """
        set up our client for the tests
        """
        cls._sc = provisioning.ShopConfigService(
            server=os.environ['EP_SERVER'],
            provider=os.environ['EP_PROVIDER'],
            username=os.environ['EP_USERNAME'],
            password=os.environ['EP_PASSWORD'],
        )
        # used as alias for these tests
        cls._nowstr = datetime.now().strftime('%Y%m%d%H%M%S%f')
        cls._shopalias_min = 'test-{}-min'.format(cls._nowstr)
        cls._shoptype = 'MinDemo'

    def test_000_get_all_info(self):
        """ get info of all shops for this provider """
        self.assertTrue(self._sc.get_all_info())

    def test_010_create_mindata(self):
        """ create shop with minimal data """
        shop = self._sc.get_createshop_obj()
        shop.Alias = self._shopalias_min
        shop.ShopAlias = self._shopalias_min
        shop.ShopType = self._shoptype
        self.assertTrue(self._sc.create(shop))

    def test_020_exists(self):
        """ check that our creation was succesfull by checking
        that the shop exists """
        shopref = self._sc.get_shopref_obj()
        shopref.Alias = self._shopalias_min
        self.assertTrue(self._sc.exists(shopref))

    def test_030_getinfo(self):
        """ fetch all the information of our shop """
        infoshop = self._sc.get_infoshop_obj()
        infoshop.Alias = self._shopalias_min
        self.assertTrue(self._sc.get_info(infoshop))

    def test_040_update(self):
        """ and update information in our shop """
        shop = self._sc.get_updateshop_obj()
        shop.Alias = self._shopalias_min
        shop.IsTrial = False
        self.assertIsNone(self._sc.update(shop))

    def test_050_set_secondary_domains(self):
        """ set some secondary domains for the shop """
        shopref = self._sc.get_shopref_obj()
        shopref.Alias = self._shopalias_min
        domains = self._sc.get_secondarydomains_obj([
            self._shopalias_min+'-test1.fi',
            self._shopalias_min+'-test2.fi'
        ])
        self.assertIsNone(self._sc.set_secondary_domains(shopref, domains))

    def test_900_delete(self):
        """ and delete our test shop """
        shopref = self._sc.get_shopref_obj()
        shopref.Alias = self._shopalias_min
        self.assertIsNone(self._sc.delete(shopref))

    def test_910_deleteref(self):
        """ and delete the shopref also """
        shopref = self._sc.get_shopref_obj()
        shopref.Alias = self._shopalias_min
        self.assertIsNone(self._sc.delete_shopref(shopref))


@envcheck
class TestSimpleProvisioning(unittest.TestCase):
    """ Tests for `epages_provisioning` package.
    All tests require that the environment variables:

        EP_SERVER
        EP_PROVIDER
        EP_USERNAME
        EP_PASSWORD

    are set

    For example:

        export EP_SERVER=http://example.com/
        export EP_PROVIDER=Distributor
        export EP_USERNAME=admin
        export EP_PASSWORD=admin
        make test

    These tests also assume that the ePages service is a default installation
    with default shoptypes etc.

    Also if you want to see a full message trace say:

        export EP_TRACE=1

    and then run your tests

    Warning: these will take a while to run.
    """

    @classmethod
    def setUpClass(cls):
        """
        set up our client for the tests
        """
        cls._sp = provisioning.SimpleProvisioningService(
            server=os.environ['EP_SERVER'],
            provider=os.environ['EP_PROVIDER'],
            username=os.environ['EP_USERNAME'],
            password=os.environ['EP_PASSWORD'],
        )
        # used as alias for these tests
        cls._nowstr = datetime.now().strftime('%Y%m%d%H%M%S%f')
        cls._shopalias_min = 'test-{}-min'.format(cls._nowstr)
        cls._shopalias_min_tmp = 'test-{}-min_tmp'.format(cls._nowstr)
        cls._shopalias_add = 'test-{}-add'.format(cls._nowstr)

        # this is probably the smallest shoptype for testing, others might be
        # "EBiz5" or "eCMSFree"
        cls._shoptype = "MinDemo"

    def test_000_create_mindata(self):
        """
        test creating new shop with minimal data
        """
        shop = self._sp.get_createshop_obj(
            {
                'Alias': self._shopalias_min,
                'ShopType': self._shoptype,
            }
        )
        self.assertIsNone(self._sp.create(shop))

    def test_001_create_missingdata(self):
        """
        test creating with missing data
        """
        # ShopType is mandatory
        shop = self._sp.get_createshop_obj(
            {
                'Alias': self._shopalias_min,
            }
        )
        with self.assertRaises(ValidationError) as e:
            self._sp.create(shop)

        self.assertEqual(e.exception.message, "Missing element ShopType")

    def test_002_create_with_additional(self):
        """ create shop with additional data, you can push any data
        you want into the shop in the additionalattributes """
        shop = self._sp.get_createshop_obj(
            {
                'Alias': self._shopalias_add,
                'ShopType': self._shoptype,
                'IsClosed': 1,
                'IsTrialShop': 1,
                'IsInternalTestShop': 1,
                # 'HasSSLCertificate': 0,
                'MerchantLogin': 'test',
                'MerchantPassword': '123456',
                'MerchantEMail': 'test@example.com',
                'Name': 'TestShop',
                'AdditionalAttributes': [
                    {
                        'Name': 'Channel',
                        'Type': 'String',
                        'Value': 'TestScript'
                    },
                    {
                        'Name': 'Likeability',
                        'Type': 'Integer',
                        'Value': '1'
                    },
                ],
            }
        )
        self.assertIsNone(self._sp.create(shop))
#        shop = self._sp.get_shopref_obj(
#            {
#                'Alias': self._shopalias_add,
#            }
#        )
#        self.assertTrue(self._sp.get_info(shop))

    def test_010_exists(self):
        """
        test exists method, assumes that shop with alias "NotExistingAlias
        does not exists in the ePages system
        """
        shop = self._sp.get_shopref_obj(
            {
                'Alias': self._shopalias_min,
            }
        )

        self.assertTrue(self._sp.exists(shop))

        shop = self._sp.get_shopref_obj(
            {
                'Alias': 'NotExistingAlias',
            }
        )
        self.assertFalse(self._sp.exists(shop))

    def test_020_get_info(self):
        """
        test getinfo
        """
        shop = self._sp.get_shopref_obj(
            {
                'Alias': self._shopalias_min,
            }
        )

        info = self._sp.get_info(shop)
        self.assertTrue(info)
        self.assertDictContainsSubset(
            {
                'Alias': self._shopalias_min,
                'MerchantLogin': None,
                'IsClosed': False,
            },
            info
        )

    def test_030_update(self):
        """
        Test updating shop data
        """
        update = self._sp.get_updateshop_obj(
            {
                'Alias': self._shopalias_min,
                'IsClosed': 1,
                'Name': "Test shop updated"
            }
        )
        self.assertIsNone(self._sp.update(update))

    def test_040_rename(self):
        """
        test rename operation (twice to get the name back to what it was)
        """
        rename = self._sp.get_rename_obj(
            {
                'Alias': self._shopalias_min,
                'NewAlias': self._shopalias_min_tmp,
            }
        )
        self.assertIsNone(self._sp.rename(rename))

        # and rename back for others
        rename = self._sp.get_rename_obj(
            {
                'Alias': self._shopalias_min_tmp,
                'NewAlias': self._shopalias_min,
            }
        )
        self.assertIsNone(self._sp.rename(rename))

    def test_900_mark_for_delete_not(self):
        """
        test shop deletion with non existing shop
        """
        shop = self._sp.get_shopref_obj(
            {
                'Alias': 'NotExistingAlias',
            }
        )
        self.assertFalse(self._sp.mark_for_deletion(shop))

    def test_901_mark_for_delete(self):
        """
        test shop deletion, assumes that creates were successfull
        """
        shop = self._sp.get_shopref_obj(
            {
                'Alias': self._shopalias_min,
            }
        )
        self.assertIsNone(self._sp.mark_for_deletion(shop))

        shop = self._sp.get_shopref_obj(
            {
                'Alias': self._shopalias_add,
            }
        )
        self.assertIsNone(self._sp.mark_for_deletion(shop))



# -*- coding: utf-8 -*-
"""
ePages provisioning service

"""
import logging

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from requests import Session
from requests.auth import HTTPBasicAuth
from zeep import Client
from zeep.transports import Transport
from zeep import Plugin

logger = logging.getLogger(__name__)


class ArrayFixer(Plugin):
    """ try to fix the soap arrays for Soap::Lite

    see https://github.com/mvantellingen/python-zeep/issues/521
    """

    def ingress(self, envelope, http_headers, operation):
        return envelope, http_headers

    def egress(self, envelope, http_headers, operation, binding_options):
        """ force array type to SecondaryDomains element

        TODO: There must be a better way
        """
        secondarydomains = envelope.find(".//SecondaryDomains")
        if secondarydomains is not None:
            logger.debug("Mangling secondarydomains element, adding arraytype")
            length = len(secondarydomains)
            secondarydomains.attrib[
                "{http://schemas.xmlsoap.org/soap/encoding/}arrayType"
            ] = "ns1:string[{}]".format(length)

            logger.debug("And removing xsitype from items")
            for item in secondarydomains.getchildren():
                item.attrib.clear()

        additional = envelope.find(".//AdditionalAttributes")
        if additional is not None:
            logger.debug("Mangling secondarydomains element, adding arraytype")
            length = len(additional)
            additional.attrib[
                "{http://schemas.xmlsoap.org/soap/encoding/}arrayType"
            ] = "ns1:anyType[{}]".format(length)

            logger.debug("And removing xsitype from items")
            for item in additional.getchildren():
                item.attrib.clear()

        return envelope, http_headers


class BaseProvisioningService(object):
    """ Base for provisioning services

    :param endpoint: server to use e.g. https://example.com/
    :param provider: provider name
    :param username: username
    :param password: password
    :param version: wsdl version number, defaults to latest version available
    """

    def __init__(
            self,
            server="",
            provider="",
            username="",
            password="",
            version=""):

        for key, value in locals().items():
            if value is "":
                raise ValueError('Argument %s required', key)

        super(BaseProvisioningService, self).__init__()

        self.server = server
        self._add_scheme_to_server()

        self.provider = provider
        self.username = username
        self.password = password
        self.version = version

        self.endpoint = self._build_endpoint_from_server()
        self.wsdl = self._build_wsdl_url_from_endpoint()
        self.userpath = self._build_full_username()

        # plugin for fixing the arrays
        arrayfixer = ArrayFixer()

        # initialize our client using basic auth and with the wsdl file
        session = Session()
        session.auth = HTTPBasicAuth(self.userpath, self.password)
        client = Client(
            wsdl=self.wsdl,
            strict=False,  # ePages wsdl files are full of errors...
            transport=Transport(session=session),
            plugins=[arrayfixer]
        )
        self.client = client

        # get the binding name, there is only one so this should be ok
        qname = next(iter(client.wsdl.bindings))

        # and create new service with the name pointing to our endpoint
        self.service2 = client.create_service(qname, self.endpoint)
        logger.debug('Initialized new client: %s', self.client)

    def _add_scheme_to_server(self):
        """ adds https:// to server if it is not there already """
        parsed = urlparse(self.server)
        if parsed.scheme == "":
            self.server = "https://{}".format(self.server)

    def _build_endpoint_from_server(self):
        """ Build endpoint url from server """
        return "{}/epages/Site.soap".format(self.server)

    def _build_wsdl_url_from_endpoint(self):
        """ you need to implement this method in subclasses, each service has
        different wsdl file locations """
        raise NotImplementedError

    def _build_full_username(self):
        """ build the username as a ePages object path """
        return "/Providers/{}/Users/{}".format(self.provider, self.username)


class ShopConfigService(BaseProvisioningService):
    """
    ShopConfig service, handles more than simple provisioning service
    """

    def __init__(self,
                 server="",
                 provider="",
                 username="",
                 password="",
                 version="11"):  # TODO: 12 after ePages fixes AD-8535
        super(ShopConfigService, self).__init__(
            server=server,
            provider=provider,
            username=username,
            password=password,
            version=version,
        )

    def _build_wsdl_url_from_endpoint(self):
        """ Builds url to the wsdl from endpoint and version number """
        parsed = urlparse(self.endpoint)
        wsdlurl = '{uri.scheme}://{uri.netloc}/WebRoot/WSDL/'\
                  'ShopConfigService{version}.wsdl'.format(
                      uri=parsed, version=self.version
                  )
        logger.debug('Built wsdl url from endpoint: %s', wsdlurl)
        return wsdlurl

    def get_infoshop_obj(self, data=None):
        """ infoshop object, used with get_info """
        if data is None:
            data = {}
        return self.client.get_type('ns0:TInfoShop_Input')(**data)

    def get_shopref_obj(self, data=None):
        """ returns a shopref object
        use this when calling exists, delete etc. """
        if data is None:
            data = {}
        return self.client.get_type('ns0:TShopRef')(**data)

    def get_all_info(self):
        """ Get info about all shops """
        return self.service2.getAllInfo()

    def get_createshop_obj(self, data=None):
        """ createshop obj
        use this when calling create """
        if data is None:
            data = {}
        return self.client.get_type('ns0:TCreateShop')(**data)

    def get_updateshop_obj(self, data=None):
        """ createshop obj
        use this when calling create """
        if data is None:
            data = {}
        return self.client.get_type('ns0:TUpdateShop')(**data)

    def get_secondarydomains_obj(self, domains):
        """ get secondarydomains obj, used with set_secondary_domains """
        if not isinstance(domains, list):
            raise TypeError(
                "domains should be a list of domains"
            )
        return self.client.get_type('ns0:TSecondaryDomains')(domains)

    def get_info(self, shop):
        """ get information about one shop

        sc.get_info(sc.get_infoshop_obj({'Alias': 'DemoShop'}))
        """
        if not isinstance(shop, type(self.get_infoshop_obj())):
            raise TypeError(
                "Get shop from get_infoshop_obj and call with that")

        return self.service2.getInfo(shop)

    def exists(self, shop):
        """ Check if a shop exists

        sc.exists(sc.get_shopref_obj({'Alias': 'DemoShop'}))
        """
        if not isinstance(shop, type(self.get_shopref_obj())):
            raise TypeError(
                "Get shop from get_shopref_obj and call with that")

        return self.service2.exists(shop)

    def create(self, shop):
        """ create new shop

        shop = sc.get_createshop_obj()
        shop.Alias = "Test"
        shop.ShopType = "MinDemo"
        sc.create(shop)
        """
        if not isinstance(shop, type(self.get_createshop_obj())):
            raise TypeError(
                "Get shop from get_createshop_obj and call with that")

        return self.service2.create(shop)

    def update(self, shop):
        """ update shop

        shop = sc.get_updateshop_obj()
        shop.Alias = "Test"
        shop.IsTrial = False
        sc.update(shop)
        """
        if not isinstance(shop, type(self.get_updateshop_obj())):
            raise TypeError(
                "Get shop from get_updateshop_obj and call with that")

        return self.service2.update(shop)

    def set_secondary_domains(self, shop, domains):
        """ set secondary domains for the shop

        shopref = sc.get_shopref_obj()
        shopref.Alias = 'DemoShop'
        domains = sc.get_secondarydomains_obj(['test.fi', 'test2.fi'])
        sc.set_secondary_domains(shopref, domains)
        """
        if not isinstance(shop, type(self.get_shopref_obj())):
            raise TypeError(
                "Get shop from get_shopref_obj and call with that")
        if not isinstance(domains, type(self.get_secondarydomains_obj([]))):
            raise TypeError(
                "Get shop from get_secondarydomains_obj and call with that")

        return self.service2.setSecondaryDomains(shop, domains)

    def delete(self, shop):
        """ delete a shop

        sc.delete(sc.get_shopref_obj({'Alias': 'DemoShop'}))
        """
        if not isinstance(shop, type(self.get_shopref_obj())):
            raise TypeError(
                "Get shop from get_shopref_obj and call with that")

        return self.service2.delete(shop)

    def delete_shopref(self, shop):
        """ delete a shop

        sc.delete_shopref(sc.get_shopref_obj({'Alias': 'DemoShop'}))
        """
        if not isinstance(shop, type(self.get_shopref_obj())):
            raise TypeError(
                "Get shop from get_shopref_obj and call with that")

        return self.service2.deleteShopRef(shop)


class SimpleProvisioningService(BaseProvisioningService):
    """ Simple provisioning

    handles creating updating and deleting shops in ePages environment.

    By default uses the following wsdl file from ePages:

    https://example.com/WebRoot/WSDL/SimpleProvisioningService6.wsdl

    You can also define the version to get other versions of the wsdl.

    The wsdl location is built from endpoint information.
    """

    def __init__(self,
                 server="",
                 provider="",
                 username="",
                 password="",
                 version="6"):
        super(SimpleProvisioningService, self).__init__(
            server=server,
            provider=provider,
            username=username,
            password=password,
            version=version
        )

    def _build_wsdl_url_from_endpoint(self):
        """ Builds url to the wsdl from endpoint and version number """
        parsed = urlparse(self.endpoint)
        wsdlurl = '{uri.scheme}://{uri.netloc}/WebRoot/WSDL/'\
                  'SimpleProvisioningService{version}.wsdl'.format(
                      uri=parsed, version=self.version
                  )
        logger.debug('Built wsdl url from endpoint: %s', wsdlurl)
        return wsdlurl

    def get_createshop_obj(self, data=None):
        """ get shop object for creation """
        if data is None:
            data = {}
        return self.client.get_type('ns0:TCreateShop')(**data)

    def get_shopref_obj(self, data=None):
        """ get shop object for exists, getinfo and markfor deletion calls """
        if data is None:
            data = {}
        return self.client.get_type('ns0:TShopRef')(**data)

    def get_updateshop_obj(self, data=None):
        """ get shop object for update call """
        if data is None:
            data = {}
        return self.client.get_type('ns0:TUpdateShop')(**data)

    def get_rename_obj(self, data=None):
        """ get rename object """
        if data is None:
            data = {}
        return self.client.get_type('ns0:TRename_Input')(**data)

    def create(self, shop):
        """ Creates new shop

        sp = provisioning.SimpleProvisioningService(...)
        shop = sp.factory_createshop_obj()
        shop.Alias = "TestShop"
        shop.ShopType = "MinDemo"
        sp.create(shop)

        returns None when everything is ok.
        """
        if not isinstance(shop, type(self.get_createshop_obj())):
            raise TypeError(
                "Get shop from get_createshop_obj and call with that")

        logger.info('Creating new shop with data: %s', shop)
        return self.service2.create(shop)

    def exists(self, shop):
        """ Check if shop exists

        shopref = sp.get_shopref_obj()
        shopref.Alias = "ExistingShop"
        exists = sp.get_info(shopref)

        """
        if not isinstance(shop, type(self.get_shopref_obj())):
            raise TypeError("Get shop from get_shopref_obj and call with that")

        return self.service2.exists(shop)

    def get_info(self, shop):
        """ Get shop information

        shopref = sp.get_shopref_obj()
        shopref.Alias = "ExistingShop"
        info = sp.get_info(shopref)

        """
        if not isinstance(shop, type(self.get_shopref_obj())):
            raise TypeError(
                "Get shop from get_shopref_obj and call with that")

        return self.service2.getInfo(shop)

    def mark_for_deletion(self, shop):
        """ Mark the shop for deletion

        The shop will be deleted by ePages after some time (default 30 days)

        Does a extra exists call before marking for deletion because
        ePages service always returns None for this call

        deleteshop = sp.get_shopref_obj()
        deleteshop.Alias = "ExistingShop"
        sp.delete(deleteshop)

        returns None if the operation was successfull
        """
        if not isinstance(shop, type(self.get_shopref_obj())):
            raise TypeError(
                "Get shop from get_shopref_obj and call with that")

        if self.service2.exists(shop):
            return self.service2.markForDeletion(shop)
        else:
            return False

    def rename(self, shop):
        """ Rename shop

        Warning: this might change some urls in the shop and confuse
        search engines etc.

        renameshop = sp.get_rename_obj()
        renameshop.Alias = "ExistingAlias"
        renameshop.NewAlias = "NewAlias"
        sp.rename(renamshop)

        returns None if the operation was successfull
        """
        if not isinstance(shop, type(self.get_rename_obj())):
            raise TypeError(
                "Get shop from get_rename_obj and call with that")

        return self.service2.rename(shop)

    def update(self, shop):
        """ Update shop information.

        updateshop = sp.get_updateshop_obj()
        updateshop.Alias = "ExistingAlias"
        updateshop.IsClosed = 1
        sp.update(updateshop)

        returns None if the operation was successfull
        """
        if not isinstance(shop, type(self.get_updateshop_obj())):
            raise TypeError(
                "Get shop from get_updateshop_obj and call with that")

        return self.service2.update(shop)
